#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# Ingo Melzer
#------------------------------------------------------------------------------
package TBD::NIS;

use Carp;
use Net::NIS ':all';
use Net::NIS::Table;
use TBI;
use strict;
use warnings;

our @ISA = qw(TBI);
our $VERSION = "0.02";

#------------------------------------------------------------------------------

=head1 NAME

TBD::NIS -- physical representation of tables as NIS maps

=head1 SYNOPSIS

   use TBD::NIS;

=head1 DESCRIPTION

TBD::NIS provides loading operations for tables which are represented as
NIS maps (as opposed to ordinary UNIX files or other representations).
TBD::NIS supports the full interface of the abstract TBI module.

Modules which are derived from TBD::NIS are expected
to invoke TBD::NIS::initialize for each new object
with following named parameters, which are all mandatory:

=over 4

=item mapname

The name of the NIS map where the table is to be taken from.

=item ifs

The input field separator which is to be given as
regular expression.

=item tablename

A string specifying the table name.

=item keyfields

A pointer to a list of key field names.

=item fields

A pointer to a list of field names. The order must reflect the
order of fields in the input file.

=back

The registration of a set of NIS maps could be arranged as in following
example:

   my %table = (
      "passwd.byname" => {
	 ifs => ':',
	 keyfields => [qw(login)],
	 fields =>
	    [qw(login passwd uid gid name home shell)],
      },
   );

   foreach my $table (keys %table) {
      TBI->register($table, 'TBD::NIS', {
	 mapname => $table,
	 tablename => $table,
	 %{$table{$table}},
      });
   }

   my $pwd = TBI->open('passwd.byname');
   print "fields:", join(" ", $pwd->fields), "\n";
   print "load:", $pwd->load, "\n";
   my %vals = $pwd->fetch('melzer');
   foreach my $val (keys %vals) { print " $val: $vals{$val}.\n"; }
   print "select gid=200 ",
      join(" ", $pwd->select(gid => 200)), ".\n";

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/melzer/">Ingo Melzer</A>

=for text
Ingo Melzer

=for man
Ingo Melzer

=cut


sub initialize1 {
   my ($self, %attributes) = @_;
   $self->SUPER::initialize1(%attributes);
   foreach my $field (qw(mapname ifs tablename fields keyfields)) {
      croak "Undefined $field" unless defined $attributes{$field};
      $self->{$field} = $attributes{$field};
   }
   my $mapname = $self->{mapname};
   my $map = Net::NIS::Table->new($mapname);
   croak "NIS map not found: $mapname" unless defined $map;
   $self->{map} = $map;
   $self->{loaded} = 0; $self->{locked} = 0;
}

sub initialize2 {
   my ($self) = @_;
   croak "Key fields not defined" unless ref($self->{keyfields});
   $self->{valid_fieldnames} = {};
   foreach my $field (@{$self->{fields}}) {
      ${$self->{valid_fieldnames}}{$field} = 1;
   }
   # check that all key field names are valid
   foreach my $keyfield (@{$self->{keyfields}}) {
      croak "Unknown field name: $keyfield"
         unless defined($self->{valid_fieldnames}->{$keyfield});
   }
   $self->SUPER::initialize2();
}

sub load {
   my $self = shift;
   my $data = $self->{map}->list();
   return 0 unless $self->{map}->status == YPERR_SUCCESS();
   my $ifs = $self->{ifs};
   my @fieldnames = $self->fields;
   while (my($key, $val) = each(%{$data})) {
      my %fields = ();
      @fields{@fieldnames} = split(/$ifs/, $val);
      $self->{'table'}->{$key} = \%fields;
   }
   $self->{loaded} = 1;
   return 1;
}

sub isloaded {
   my $self = shift;
   return $self->{loaded};
}

sub name {
   my $self = shift;
   return $self->{tablename};
}

sub fields {
   my $self = shift;
   return @{$self->{fields}};
}

sub keyfields {
   my $self = shift;
   return @{$self->{keyfields}};
}

sub keys {
   my $self = shift;
   return CORE::keys %{$self->{table}};
}

sub exists {
   my ($self, $key) = @_;
   return exists $self->{table}->{$key};
}

sub get {
   my ($self, $key, @fieldnames) = @_;
   my %record = $self->fetch($key);
   return @record{@fieldnames};
}

sub getfield {
   my ($self, $key, $fieldname) = @_;
   return scalar $self->get($key, $fieldname);
}

sub select {
   my ($self, %equations) = @_;
   if (%equations) {
      my @keys = ();
      LOOP: while (my($key, $fields) = each(%{$self->{table}})) {
         my %fields = %{$fields};
         foreach my $fieldname (CORE::keys %equations) {
            next LOOP if $fields{$fieldname} ne $equations{$fieldname}
         }
         push(@keys, $key);
      }
      return @keys;
   } else {
      # no equations given, so return all keys
      return $self->keys;
   }
}

sub add {
   my $self = shift;
   croak "Trying to modify read-only table :-(";
   return 0;
}

sub delete {
   my $self = shift;
   croak "Trying to modify read-only table :-(";
   return 0;
}

sub modify {
   my $self = shift;
   croak "Trying to modify read-only table :-(";
   return 0;
}

sub close {
   #my $self = shift;
#   $self->SUPER::close();
}

sub fetch {
   my ($self, $key) = @_;
   croak "no such key" unless defined $self->{table}->{$key};
   return %{$self->{table}->{$key}};
}

1;
