#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 2/96, etter 2/97, afb 12/1999
#------------------------------------------------------------------------------
# implementation of a table as ordinary text file
#------------------------------------------------------------------------------

package TBD::Text;
use Carp;
use English;
use Fcntl;
use File::Sync;
use IO::File;
use TBD::CachingTable;
use TBD::Dumper;
use TBD::Locker;
use TBD::Scanner;
use TBI;
use strict;
use warnings;

our $VERSION = "0.03";
our @ISA = qw(TBD::CachingTable TBD::Scanner TBD::Locker TBD::Dumper);

=head1 NAME

TBD::Text - physical representation of tables as ordinary text files

=head1 SYNOPSIS

   use TBD::Text;
   our @ISA = qw(... TBD::Text ...);

=head1 DESCRIPTION

TBD::Text provides loading and storing operations for tables which are
represented as ordinary text files (as opposed to DBM files, NIS maps or
database tables).  Because TBD::Text is an extension of TBD::CachingTable
all accessing operations are provided as well.

Following parameters are expected to be passed to constructors.
Modules which are derived from TBD::Text are expected to pass these
parameters to TBD::Text::initialize1.

=over 4

=item filename

The database file where the table is to be loaded from and
to be stored to.

=item tmpfile

The name of a temporary file that is used to create a new
text file version. This file is renamed to I<filename> by
an atomic operation. For this to work, I<tmpfile> must reside
on the same file system as I<filename>.

=item logfile

The name of a logfile where records about changes are to be added.

=item ifs

The input field separator which is to be given as
regular expression.

=item irs

The input record separator which is to be given as string.

=item ofs

The output field separator which is to be used on storing
modificated tables back to the original file.

=item ors

The output record separator -- usually identical to the
input record separator.

=item comment

A string which (or regular expression without leading ^) which matches
comment lines which are to be ignored in the file.  This string may be
empty if no comments are used.

=item tablename

A string specifying the table name.

=item keyfields

A pointer to a list of key field names.

=item fields

A pointer to a list of field names. The order must reflect the
order of fields in the input file.

=back

The parameters I<fields>, I<filename>, I<ifs>, I<irs>, and I<tablename>
are mandatory. Reasonable defaults are provided for I<keyfields>
(first field name of I<fields>), I<ofs> (same as I<ifs>), I<ors> (same
as I<irs>), I<tmpfile> (I<filename> with ``.tmp'' appended), I<logfile>
(I<filename> with ``.log'' appended), and I<comment> (empty string).

The registration of a set of text file tables could
be arranged as in following example:

   my %table = (
      Members => {
         ifs => ':',
         irs => "\n",
         keyfields => [qw(login)],
         fields => [qw(login name)],
      },
      Documents => {
         ifs => '~',
         irs => "\n",
         keyfields => [qw(group shortname)],
         fields => [qw(group shortname login title)],
      },
      # ...
   );

   my $tabledir = "/some/path/where/your/tables/reside";
   foreach my $table (keys %table) {
      TBI->register($table, 'TBD::Text', {
	 filename => "$tabledir/$table",
	 tablename => $table,
	 %{$table{$table}},
      });
   }

Note that TBD::Text is an extension of TBD::Scanner, TBD::CachingTable,
TBD::Dumper, and TBD::Locker. In addition, TBD::Text provides following
methods:

=over 4

=item $table->load

loads (or re-loads) the table from the current file representation.
This operation is required for read accesses.

=item $table->isloaded

returns true if reading operations are permitted.

=item $table->lock

locks the table for write operations. This lock will be held exclusively
and will block all other parties that are using this implementation and
try to lock the table as well.

=item $table->unlock

saves the current state of the table back to the associated file and
unlocks it. Read-only accesses are still permitted afterwards.

=item $table->islocked

returns true if writing operations are permitted.

=item $table->valid($field)

returns true if the contents of $field does not interfere with
field and record separators, i.e. if it is save to be assigned
to a field.

=back

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>
with contributions of
<A HREF="http://www.mathematik.uni-ulm.de/sai/etter/">Mathias Etter</A>.

=for text
Andreas Borchert with contributions of Mathias Etter.

=for man
Andreas Borchert with contributions of Mathias Etter.

=cut

sub initialize1 {
   my ($self, %attributes) = @_;

   $self->SUPER::initialize1(%attributes);

   foreach my $field (qw(filename ifs irs tablename fields)) {
      croak "Undefined $field" unless defined $attributes{$field};
      $self->{$field} = $attributes{$field};
   }
   croak "Pointer to list of field names expected"
      unless ref($attributes{fields}) eq "ARRAY";
   my %defaults = (
      ofs => $attributes{ifs},
      ors => $attributes{irs},
      tmpfile => $attributes{filename} . ".tmp",
      logfile => $attributes{filename} . ".log",
      keyfields => [qw($attributes{fields}->[0])],
      comment => "",
   );
   foreach my $field (keys %defaults) {
      if (defined $attributes{$field}) {
	 $self->{$field} = $attributes{$field};
      } else {
	 $self->{$field} = $defaults{$field};
      }
   }
   my $filename = $self->{filename};
   croak "Table file not found: $filename" unless -r $filename;
   $self->{loaded} = 0; $self->{locked} = 0;
   $self->{lockfh} = undef; $self->{outfh} = undef;
}

sub load {
   my $self = shift;

   my $tab = new IO::File $self->{filename};
   return 0 unless defined $tab;
   my $ok = $self->scan($tab);
   $tab->close;
   return 0 unless $ok;
   $self->{loaded} = 1;
   return 1;
}

sub isloaded {
   my $self = shift;
   return $self->{loaded};
}

sub lock {
   my $self = shift;
   my $tmpfile = $self->{tmpfile};
   my $filename = $self->{filename};

   $self->unlock() if $self->{locked};

   my $in = new IO::File $filename, O_RDWR
      or croak "unable to open $filename for reading and writing: $!";
   unless ($self->SUPER::lock($in) && $self->scan($in)) {
      $in->close;
      return 0;
   }
   my $out = new IO::File $tmpfile, O_WRONLY|O_CREAT|O_EXCL, 0600;
   unless (defined $out) {
      unlink($tmpfile);
      $out = new IO::File $tmpfile, O_WRONLY|O_CREAT|O_EXCL, 0600;
      unless (defined $out) {
	 $in->close;
	 return 0;
      }
   }

   $self->{lockfh} = $in;
   $self->{outfh} = $out;
   $self->{loaded} = 1;
   $self->{locked} = 1;
   return 1;
}

sub unlock {
   my $self = shift;
   my $filename = $self->{filename};
   my $tmpfile = $self->{tmpfile};
   my $lockfh = $self->{lockfh};
   my $outfh = $self->{outfh};

   return 0 unless $self->{locked} && defined $lockfh && defined $outfh;

   my $ok = $self->dump($outfh) &&
	    $outfh->flush() && $outfh->fsync() && $outfh->close() &&
	    rename($tmpfile, $filename);

   $self->SUPER::unlock($lockfh);
   $lockfh->close();
   $self->{locked} = 0;
   $self->{lockfh} = undef;
   $self->{outfh} = undef;

   return $ok;
}

sub islocked {
   my $self = shift;
   my $fh = $self->{lockfh};

   return $self->{locked} || $self->SUPER::islocked($fh)
}

sub name {
   my $self = shift;
   return $self->{tablename};
}

sub fields {
   my $self = shift;
   return @{$self->{fields}};
}

sub keyfields {
   my $self = shift;
   return @{$self->{keyfields}};
}

sub ofs {
   my $self = shift;

   return $self->{ofs};
}

sub valid {
   my ($self, $field) = @_;

   return 0 if $field =~ /$self->{ifs}/;
   return 0 if index($field, $self->{irs}) >= 0;

   return 0 if index($field, $self->{ofs}) >= 0;
   return 0 if index($field, $self->{ors}) >= 0;
   return 1;
}

sub close {
   my $self = shift;
   $self->unlock() if $self->{locked};
   $self->SUPER::close();
}

1;
